import unittest
import os, os.path, glob
from SMART.Java.Python.structure.Transcript import Transcript
from SMART.Java.Python.GetFlanking import GetFlanking
from commons.core.writer.Gff3Writer import Gff3Writer
from commons.core.parsing.GffParser import GffParser

class Test_F_GetFlanking(unittest.TestCase):

    def setUp(self):
        self.queryFileName     = "testQuery.gff3"
        self.referenceFileName = "testReference.gff3"
        self.outputFileName    = "testOutput.gff3"
         
    def tearDown(self):
        for fileRoot in (self.queryFileName, self.referenceFileName, self.outputFileName):
            for file in glob.glob("%s*" % (fileRoot)):
                os.remove(file)

    def test_run_simple(self):
        reference = self._createTranscript("chr1", 1000, 1100, "+", "ref")
        writer = Gff3Writer(self.referenceFileName, 0)
        writer.addTranscript(reference)
        writer.close()
        query1 = self._createTranscript("chr1", 100, 200, "+", "query1")
        query2 = self._createTranscript("chr1", 10000, 10100, "+", "query2")
        writer = Gff3Writer(self.queryFileName, 0)
        writer.addTranscript(query1)
        writer.addTranscript(query2)
        writer.close()
        gf = GetFlanking(0)
        gf.setInputFile(self.queryFileName, 'gff3', 0)
        gf.setInputFile(self.referenceFileName, 'gff3', 1)
        gf.setOutputFile(self.outputFileName)
        gf.run()
        parser = GffParser(self.outputFileName)
        self.assertEqual(parser.getNbTranscripts(), 1)
        for transcript in parser.getIterator():
            self._checkTranscript(transcript, "chr1", 100, 200, "+", "query1")

    def test_run_simple_downstream(self):
        reference = self._createTranscript("chr1", 1000, 1100, "+", "ref")
        writer = Gff3Writer(self.referenceFileName, 0)
        writer.addTranscript(reference)
        writer.close()
        query1 = self._createTranscript("chr1", 100, 200, "+", "query1")
        query2 = self._createTranscript("chr1", 1200, 1300, "+", "query2")
        query3 = self._createTranscript("chr1", 1400, 1500, "+", "query3")
        writer = Gff3Writer(self.queryFileName, 0)
        writer.addTranscript(query1)
        writer.addTranscript(query2)
        writer.addTranscript(query3)
        writer.close()
        gf = GetFlanking(0)
        gf.setInputFile(self.queryFileName, 'gff3', 0)
        gf.setInputFile(self.referenceFileName, 'gff3', 1)
        gf.setOutputFile(self.outputFileName)
        gf.addDownstreamDirection(True)
        gf.run()
        parser = GffParser(self.outputFileName)
        self.assertEqual(parser.getNbTranscripts(), 1)
        for transcript in parser.getIterator():
            self._checkTranscript(transcript, "chr1", 1200, 1300, "+", "query2")

    def test_run_simple_minus_strand_downstream(self):
        reference = self._createTranscript("chr1", 1000, 1100, "-", "ref")
        writer = Gff3Writer(self.referenceFileName, 0)
        writer.addTranscript(reference)
        writer.close()
        query1 = self._createTranscript("chr1", 100, 200, "+", "query1")
        query2 = self._createTranscript("chr1", 1200, 1300, "+", "query2")
        query3 = self._createTranscript("chr1", 1400, 1500, "+", "query3")
        writer = Gff3Writer(self.queryFileName, 0)
        writer.addTranscript(query1)
        writer.addTranscript(query2)
        writer.addTranscript(query3)
        writer.close()
        gf = GetFlanking(0)
        gf.setInputFile(self.queryFileName, 'gff3', 0)
        gf.setInputFile(self.referenceFileName, 'gff3', 1)
        gf.setOutputFile(self.outputFileName)
        gf.addDownstreamDirection(True)
        gf.run()
        parser = GffParser(self.outputFileName)
        self.assertEqual(parser.getNbTranscripts(), 1)
        for transcript in parser.getIterator():
            self._checkTranscript(transcript, "chr1", 100, 200, "+", "query1")

    def test_run_simple_upstream(self):
        reference = self._createTranscript("chr1", 1000, 1100, "+", "ref")
        writer = Gff3Writer(self.referenceFileName, 0)
        writer.addTranscript(reference)
        writer.close()
        query1 = self._createTranscript("chr1", 100, 200, "+", "query1")
        query2 = self._createTranscript("chr1", 300, 400, "+", "query2")
        query3 = self._createTranscript("chr1", 1200, 1300, "+", "query3")
        writer = Gff3Writer(self.queryFileName, 0)
        writer.addTranscript(query1)
        writer.addTranscript(query2)
        writer.addTranscript(query3)
        writer.close()
        gf = GetFlanking(0)
        gf.setInputFile(self.queryFileName, 'gff3', 0)
        gf.setInputFile(self.referenceFileName, 'gff3', 1)
        gf.setOutputFile(self.outputFileName)
        gf.addUpstreamDirection(True)
        gf.run()
        parser = GffParser(self.outputFileName)
        self.assertEqual(parser.getNbTranscripts(), 1)
        for transcript in parser.getIterator():
            self._checkTranscript(transcript, "chr1", 300, 400, "+", "query2")

    def test_run_simple_colinear(self):
        reference = self._createTranscript("chr1", 1000, 1100, "+", "ref")
        writer = Gff3Writer(self.referenceFileName, 0)
        writer.addTranscript(reference)
        writer.close()
        query1 = self._createTranscript("chr1", 1200, 1300, "-", "query1")
        query2 = self._createTranscript("chr1", 1400, 1500, "+", "query2")
        writer = Gff3Writer(self.queryFileName, 0)
        writer.addTranscript(query1)
        writer.addTranscript(query2)
        writer.close()
        gf = GetFlanking(0)
        gf.setInputFile(self.queryFileName, 'gff3', 0)
        gf.setInputFile(self.referenceFileName, 'gff3', 1)
        gf.setOutputFile(self.outputFileName)
        gf.addDownstreamDirection(True)
        gf.setColinear(True)
        gf.run()
        parser = GffParser(self.outputFileName)
        self.assertEqual(parser.getNbTranscripts(), 1)
        for transcript in parser.getIterator():
            self._checkTranscript(transcript, "chr1", 1400, 1500, "+", "query2")

    def test_run_simple_max_distance(self):
        reference = self._createTranscript("chr1", 1000, 1100, "+", "ref")
        writer = Gff3Writer(self.referenceFileName, 0)
        writer.addTranscript(reference)
        writer.close()
        query1 = self._createTranscript("chr1", 2000, 2100, "-", "query1")
        writer = Gff3Writer(self.queryFileName, 0)
        writer.addTranscript(query1)
        writer.close()
        gf = GetFlanking(0)
        gf.setInputFile(self.queryFileName, 'gff3', 0)
        gf.setInputFile(self.referenceFileName, 'gff3', 1)
        gf.setOutputFile(self.outputFileName)
        gf.addDownstreamDirection(True)
        gf.setMaxDistance(100)
        gf.run()
        parser = GffParser(self.outputFileName)
        self.assertEqual(parser.getNbTranscripts(), 0)

    def _createTranscript(self, chromosome, start, end, strand, name):
        transcript = Transcript()
        transcript.setChromosome(chromosome)
        transcript.setStart(start)
        transcript.setEnd(end)
        transcript.setDirection(strand)
        transcript.setName(name)
        return transcript

    def _checkTranscript(self, transcript, chromosome, start, end, strand, name):
        self.assertEqual(transcript.getChromosome(), chromosome)
        self.assertEqual(transcript.getStart(), start)
        self.assertEqual(transcript.getEnd(), end)
        self.assertEqual(transcript.getStrand(), strand)
        self.assertEqual(transcript.getName(), name)

        
if __name__ == "__main__":
    unittest.main()
