#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import getopt
import time
import glob
import sys
import os

from commons.core.checker.CheckerException import CheckerException
from commons.core.sql.RepetJob import RepetJob
from commons.core.sql.Job import Job
from pyRepet.launcher.AbstractClusterLauncher import AbstractClusterLauncher

GENERIC_IN_FILE = "zDUMMYz"


## Abstract class to launch a program in parallel on a cluster.
#
class AbstractClusterLauncher( AbstractClusterLauncher ):  #( IClusterLauncher )
     
    def __init__( self ):
        """
        Constructor.
        """
        self._inputDir = ""            # path to the directory with input files
        self._queueName = ""           # name of the queue on the cluster
        self._groupId = ""             # identifier of the group of jobs (groupid)
        self._inputFileSuffix = "fa"   # suffix of the input files (default='fa')
        self._prgAcronym = ""          # acronym of the program to launch
        self._configFile = ""          # name of the configuration file (connect to MySQL)
        self._currentDir = os.getcwd() # path to the current directory
        self._tmpDir = ""              # path to the temporary directory
        self._jobTable = "jobs"        # name of the table recording the jobs
        self._catOutFiles = False      # concatenate output files of all jobs
        self._clean = False            # clean job file, job stdout, job table...
        self._verbose = 1              # verbosity level
        self.jobdb = None              # RepetJob instance
        self.job = Job()               # Job instance
        self._exeWrapper = "AbstractProgramLauncher.py"
        self._prgLauncher = None
        # list of instances derived from AbstractProgramLauncher()
        # If several program are launched successively in the same job,
        # 'lPrgLaunchers' has to be filled before run().
        self.lPrgLaunchers = []
    
    def setProgramLauncherAttributeFromCmdLine(self, o, a=""):
        self._getProgramLauncherInstance().setASpecificAttributeFromCmdLine(o, a)
      
    def setClusterLauncherAttributeFromCmdLine(self, o, a=""):
        if o == "-h":
            print self.getHelpAsString()
            sys.exit(0)
        elif o == "-i":
            self.setInputDirectory(a)
        elif o == "-Q":
            self.setQueueName(a)
        elif o == "-g":
            self.setGroupIdentifier(a)
        elif o == "-S":
            self.setInputFileSuffix(a)
        elif o == "-a":
            self.setAcronym(a)
        elif o == "-C":
            self.setConfigFile(a)
        elif o == "-d":
            self.setTemporaryDirectory(a)
        elif o == "-j":
            self.setJobTableName(a)
        elif o == "-Z":
            self.setCatOutputFiles()
        elif o == "-c":
            self.setClean()
        elif o == "-v":
            self.setVerbosityLevel(a)

    def setAttributesFromCmdLine(self):
        try:
            opts, args = getopt.getopt(sys.argv[1:], self.getCmdLineOptions())
        except getopt.GetoptError, err:
            print str(err);
            print self.getHelpAsString()
            sys.exit(1)
        for o, a in opts:
            self.setClusterLauncherAttributeFromCmdLine(o, a)
            self.setProgramLauncherAttributeFromCmdLine(o, a)
 
    def getHelpAsString(self):
        string = ""
        string += "usage: %s.py [options]" % (type(self).__name__ )
        string += "\ngeneric options:"
        string += "\n     -h: this help"
        string += "\n     -i: directory with input files (absolute path)"
        string += "\n     -Q: name of the queue on the cluster"
        string += "\n     -g: identifier of the group of jobs (groupid)"
        string += "\n     -S: suffix of the input files (default='fa')"
        string += "\n     -a: acronym of the program to be launched (default='%s')" % ( self.getAcronym() )
        string += "\n     -C: configuration file to connect to MySQL (absolute path or in current dir)"
        string += "\n     -d: temporary directory (absolute path, default=None)"
        string += "\n     -j: table recording the jobs (default='jobs')"
        string += "\n     -c: clean the temporary data"
        string += "\n     -v: verbosity level (default=0/1/2)"
        return string

    def formatGroupidAndTime(self):
        return self.job.groupid + " " + time.strftime("%Y-%m-%d %H:%M:%S") + ""

    def submitJob(self, lInFiles):
        count = 0
        for inFile in lInFiles:
            count += 1
            fileName = os.path.basename(inFile)
            if self._verbose > 1:
                print "processing '%s' # %i..." % (fileName, count)
                sys.stdout.flush()
            
            self.initializeJob(fileName, count)
            time.sleep(0.5)
            exitStatus = self.jobdb.submitJob(self.job)
            if exitStatus != 0:
                print "ERROR while submitting job '%i' to the cluster" % (count)
                sys.exit(1)

    def getInputFilesList(self):
        lInFiles = glob.glob("%s/*.%s" % (self._inputDir, self._inputFileSuffix))
        return lInFiles

    def checkClusterLauncherAttributes( self ):
        if self.getInputDirectory() == "":
            message = "ERROR: missing input directory"
            raise CheckerException(message)
        if not os.path.exists( self.getInputDirectory() ):
            message = "ERROR: input directory '%s' doesn't exist" % ( self.getInputDirectory() )
            raise CheckerException(message)
        if self.getGroupIdentifier() == "":
            message = "ERROR: missing group identifier"
            raise CheckerException(message)
        if self.getAcronym() == "":
            message = "ERROR: missing program acronym"
            raise CheckerException(message)
        if self.getConfigFile() == "":
            message = "ERROR: missing config file to access MySQL"
            raise CheckerException(message)
        if not os.path.exists( self.getConfigFile() ):
            message = "ERROR: config file '%s' doesn't exist" % ( self.getConfigFile() )
            raise CheckerException(message)   
        if self.getTemporaryDirectory() == "":
            self.setTemporaryDirectory(self._currentDir)
            
    def checkProgramLauncherAttributes( self ):
        self._prgLauncher.checkSpecificAttributes()

    def start( self ):
        try:
            self.checkClusterLauncherAttributes()
        except CheckerException, msg:
            print msg
            print self.getHelpAsString()
            sys.exit(1)
            
        if self.lPrgLaunchers == []:
            self.setSingleProgramLauncher()
        for pL in self.lPrgLaunchers:
            if pL.getWrapperCommandLine() == "":
                string = "ERROR: wrapper command is empty !"
                print string
                sys.exit(1)
            if pL.getProgramCommandLine() == "":
                string = "ERROR: program command is empty !"
                print string
                sys.exit(1)
        self.checkProgramAvailability()
        
        try:
            self.checkProgramLauncherAttributes()
        except CheckerException, msg:
            print msg
            print self.getHelpAsString()
            sys.exit(1)
        
        if self.getVerbosityLevel() > 0:
            string = "START %s" % ( type(self).__name__ )
            print string
        self.job.tablename = self.getJobTableName()
        self.job.groupid = self.getGroupIdentifier()
        tokens = self.getQueueName().replace("'","").split(" ")
        self.job.setQueue( tokens[0] )
        if len(tokens) > 1:
            lResources = tokens[1:]
            self.job.lResources = lResources
        if self.getVerbosityLevel() > 0:
            print "groupid: %s" % ( self.getGroupIdentifier() )
        self.jobdb = RepetJob( cfgFileName=self.getConfigFile() )
        if self.jobdb.hasUnfinishedJob( self.job.tablename, \
                                          self.job.groupid ):
            self.jobdb.waitJobGroup( self.job.tablename, self.job.groupid )
            return
        self.jobdb.cleanJobGroup( self.job.tablename, self.job.groupid )
        sys.stdout.flush()
   
    def end( self ):
        if self.getClean():
            self.removeAllJobFiles()
            self.removeAllJobStdouts()
            self.removeAllJobStderrs()
            
        if self.getCatOutputFiles():
            self.catOutputFiles()
        
        self.jobdb.close()
        
        if self.getVerbosityLevel() > 0:
            string = "END %s" % ( type(self).__name__ )
            print string
        sys.stdout.flush()
        
    def run( self ):
        #TODO: self.checkClusterLauncherAttributes() in preProcess()
        self.preProcess()
        self.start()
        
        lInFiles = self.getInputFilesList()
        
        if self._verbose > 0:
            string = "submitting " + str(len(lInFiles)) +  " jobs... " + self.formatGroupidAndTime() 
            print string; sys.stdout.flush()
        
        self.submitJob(lInFiles)
        
        if self._verbose > 0:        
            string = "waiting for jobs... " + self.formatGroupidAndTime() 
            print string; sys.stdout.flush()
        
        self.jobdb.waitJobGroup( self.job.tablename, self.job.groupid )
        
        if self._verbose > 0:
            string = "all jobs completed ! " + self.formatGroupidAndTime() 
            print string; sys.stdout.flush()
            statsExecutionTime = self.getStatsOfExecutionTime()
            print "execution time of all jobs (seconds): %f" % statsExecutionTime.getSum()
            print "execution time per job: %s" % statsExecutionTime.string()
            sys.stdout.flush()
            
        self.jobdb.cleanJobGroup( self.job.tablename, self.job.groupid )
        
        self.end()

    def getCmdLineOptions(self):
        return "hi:Q:g:S:a:C:d:j:Zcv:"
            
    def preProcess(self):
        pass
    
    def getProgramLauncherInstance(self):
        pass
